@@ -8,7 +8,7 @@
 #' On this clusterforest object, several methods, such as plot, print and summary, can be used.
 #'
 #' @param observeddata The entire observed dataset
-#' @param treedata A list of dataframes on which the trees are based. Not necessary if the data set is included in the tree object.
+#' @param treedata A list of dataframes on which the trees are based. Not necessary if the data set is included in the tree object already.
 #' @param trees A list of trees of class party, classes inheriting from party (e.g., glmtree), classes that can be coerced to party (i.e., rpart, Weka_tree, XMLnode), or a randomForest or ranger object.
 #' @param simmatrix A similaritymatrix with the similarities between all trees. Should be square, symmetric and have ones on the diagonal. Default=NULL
 #' @param m Similarity measure that should be used to calculate similarities, in the case that no similarity matrix was provided by the user. Default=NULL.
@@ -25,9 +25,11 @@
 #' @param weight If 1, the number of dissimilar paths in the Shannon and Banks measure (m=2), should be weighted by 1/their length (Otherwise they are weighted equally). Only applicable for m=2. Default=NULL
 #' @param fromclus The lowest number of clusters for which the PAM algorithm should be run. Default=1.
 #' @param toclus The highest number of clusters for which the PAM algorithm should be run. Default=1.
-#' @param treecov A vector/dataframe with the covariate value(s) for each tree in the forest (1 column per covariate).
+#' @param treecov A vector/dataframe with the covariate value(s) for each tree in the forest (1 column per covariate) in the case of known
+#' sources of variation underlying the forest, that should be linked to the clustering solution.
 #' @param sameobs Are the same observations included in every tree data set? For example, in the case of subsamples or bootstrap samples, the answer is no. Default=FALSE
 #' @param seed A seed number that should be used for the multi start procedure (based on which initial medoids are assigned). Default=NULL.
+#' @param no_cores Number of CPU cores used for computations. Default=detectCores(logical=FALSE)
 #' @return The function returns an object of class clusterforest, with attributes:
 #' \item{medoids}{the position of the medoid trees in the forest (i.e., which element of the list of partytrees)}
 #' \item{medoidtrees}{the medoid trees}
@@ -51,6 +53,8 @@
 #' @importFrom methods is
 #' @importFrom ranger treeInfo
 #' @importFrom randomForest getTree
+#' @importFrom foreach foreach %do% %dopar%
+#' @importFrom doParallel registerDoParallel
 #' @import MASS
 #' @import partykit
 #' @import rpart
@@ -61,6 +65,7 @@
 #' @references \cite{Chipman, H. A., George, E. I., & McCulloh, R. E. (1998). Making sense of a forest of trees. Computing Science and Statistics, 84-92.}
 #' @examples
 #' require(MASS)
+#' require(ranger)
 #' require(rpart)
 #'#Function to draw a bootstrap sample from a dataset
 #'DrawBoots <- function(dataset, i){
@@ -85,10 +90,17 @@
 #'
 #'#Clustering the trees in this forest
 #'ClusterForest<- clusterforest(observeddata=Pima.tr,treedata=Boots,trees=Trees,m=1,
-#'fromclus=1, toclus=3, sameobs=FALSE)
+#'fromclus=1, toclus=2, sameobs=FALSE, no_cores=2)
 #'
+#'#Example RandomForest
+#'Pima.tr.ranger <- ranger(type ~ ., data = Pima.tr, keep.inbag = TRUE, num.trees=20,
+#'max.depth=3)
+#'
+#'ClusterForest<- clusterforest(observeddata=Pima.tr,trees=Pima.tr.ranger,m=5,
+#'                            fromclus=1, toclus=2, sameobs=FALSE, no_cores=2)
 
-clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m=NULL, tol=NULL, weight=NULL,fromclus=1, toclus=1, treecov=NULL, sameobs=FALSE, seed=NULL){
+
+clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m=NULL, tol=NULL, weight=NULL,fromclus=1, toclus=1, treecov=NULL, sameobs=FALSE, seed=NULL, no_cores = detectCores(logical=FALSE)){
   ############################## Check forest input #####################
   ###  Some checks whether correct forest information is provided by user
   if(typeof(trees) != "list" ) {
@@ -129,6 +141,8 @@ clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m
     treedt=treedata
   }
   
+  cl <- makeCluster(no_cores)
+  doParallel::registerDoParallel(cl)
   
   
   ## Turn user provided forest information into object of class forest.
@@ -221,7 +235,6 @@ clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m
         cat("This measure only works on numerical splitvariables (class integer or numeric)")
         return(NULL)
       }
-
       sim <- M6(forest$observeddata, forest$treedata, Y, X, forest$partytrees, tol=NULL)
     }
 
@@ -286,44 +299,51 @@ clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m
   medsseeds<- list(0)
   correct<- list(0)
 
-  for (i in fromclus:toclus){
-
+  pamtrees <- lapply(1:length(trees), function (i) pamtree(observeddata, treedata[[i]], Y[i], trees[[i]]))
+  g <- lapply(1:length(trees), function(k) pamtrees[[k]]$predresp)
+  
+  g_matrix <- t(sapply(1:length(trees), function(k) as.numeric(levels(g[[k]])[g[[k]]])))
+  
+  forpred <- function(k) {
+    levels_g1[which.max( c(sum(g_matrix[,k] == levels_g1[1], na.rm=TRUE),  sum(g_matrix[,k] == levels_g1[2], na.rm=TRUE)))]
+  }
+  levels_g1 = levels(g[[1]])
+  forestpred <- sapply(1:nrow(observeddata), forpred)
+  
+  
+  for (i in fromclus:toclus) {
     # first do the clustering with BUILD and SWAP phase
     clustering <- pam(x = 1 - treesimilarities, k = i, diss = TRUE, pamonce=3)
 
     if(is.null(seed)){
-      seed<- round(runif(1,0,100000))
+      seed <- round(runif(1,0,100000))
     }
-    #Then try 1000 random starts + SWAP Phase
-    for (j in 1:1000){
+    #Then try 100 random starts + SWAP Phase
+    for (j in 1:100){
       set.seed(seed+j)
-      initmedoids= sample(1:nrow(treesimilarities),i)
+      initmedoids = sample(1:nrow(treesimilarities),i)
       medsseeds[[j]] <- pam(1-treesimilarities,k=i,medoids=initmedoids, diss=TRUE, pamonce=3)
-      obj[j]<-medsseeds[[j]]$objective[2]
+      obj[j] <- medsseeds[[j]]$objective[2]
     }
 
-    # check whether objective function of one of the multistarts is better than the one of the build alogrithm
+    # check whether objective function of one of the multistarts is better than the one of the build algorithm
     # and if so, continue with the best result
      if( round(min(obj), 10) < round(clustering$objective[2], 10) ){
-      clustering<- medsseeds[[  which.min(round(min(obj), 10))]]
+      clustering <- medsseeds[[  which.min(round(min(obj), 10))]]
      }
 
 
-    medoids[[i]] <- clustering $ medoids
-    clusters[[i]] <- clustering $ clustering
+    medoids[[i]] <- clustering$medoids
+    clusters[[i]] <- clustering$clustering
 
-    meds<- list(0)
+    meds <- vector(mode = "list", length = i)
     for(j in 1:i){
       meds[[j]] <- trees[[medoids[[i]][j]]]
     }
 
-    mds[[i]]<- meds
+    mds[[i]] <- meds
     sil[[i]] <-  clustering $ silinfo $ avg.width
 
-    pamtrees <- lapply(1:length(trees), function (i) pamtree(observeddata,treedata[[i]], Y[i],trees[[i]]))
-    g<- lapply(1:length(trees), function(k) pamtrees[[k]]$predresp)
-    forestpred <- sapply(1:nrow(observeddata), function (k) levels(g[[1]])[which.max( c(sum(unlist(lapply(g, `[[`, k)) ==  levels(unlist(lapply(g, `[[`, k)))[1], na.rm=TRUE),  sum(unlist(lapply(g, `[[`, k)) ==  levels(unlist(lapply(g, `[[`, k)))[2], na.rm=TRUE) )   )] )
-
     gmed <- g[c(medoids[[i]])]
     w <- table(clusters[[i]])
     
@@ -331,27 +351,35 @@ clusterforest <- function (observeddata, treedata=NULL, trees, simmatrix=NULL, m
     #unweighted
     #medpred1 <- sapply(1:nrow(observeddata), function (k) levels(gmed[[1]])[which.max( c(sum(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[1], na.rm=TRUE),  sum(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[2], na.rm=TRUE) )   )] )
     
+    #gmed_matrix <- as.numeric(levels(gmed[[1]])[gmed[[1]]])
+    gmed_matrix <- t(sapply(1:i, function(k) as.numeric(levels(gmed[[k]])[gmed[[k]]])))
+    levels_gmed1 = levels(gmed[[1]])
+    
+    medpred_weighted <- function(k) {
+      levels_gmed1[which.max( c(sum((gmed_matrix[,k] == levels_gmed1[1]) * w , na.rm=TRUE),  sum((gmed_matrix[,k] == levels_gmed1[2]) * w, na.rm=TRUE)))]
+    }
     #weighted
-    medpred <-sapply(1:nrow(observeddata), function (k) levels(gmed[[1]])[which.max( c(sum(as.numeric(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[1]) *w , na.rm=TRUE),  sum(as.numeric(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[2]) * w, na.rm=TRUE) )   )] )
-    
-    
+    medpred <- sapply(1:nrow(observeddata), medpred_weighted)
+
+
     agreement[[i]] <- mean(forestpred == medpred)
     
-    correct[[i]] <- mean(forest$observeddata[,Y][1]== medpred)
+    correct[[i]] <- mean(forest$observeddata[,Y][1] == medpred)
 
     sumw<- numeric(0)
     for (j in 1:i){
       sumw[j] <- sum(treesimilarities[clusters[[i]]==j, medoids[[i]][j]])
     }
-    sums[[i]]<- sum(sumw) / dim(treesimilarities)[1]
+    sums[[i]] <- sum(sumw) / dim(treesimilarities)[1]
+    
   }
 
-  #Accuracy of forest as a whole and of marginally best class
-  medpred <-sapply(1:nrow(observeddata), function (k) levels(gmed[[1]])[which.max( c(sum(as.numeric(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[1]) *w , na.rm=TRUE),  sum(as.numeric(unlist(lapply(gmed, `[[`, k)) ==  levels(unlist(lapply(gmed, `[[`, k)))[2]) * w, na.rm=TRUE) )   )] )
+  stopCluster(cl)
   
-  correct[[i+1]] <- mean(forest$observeddata[,Y][1]== forestpred)
-  correct[[i+2]] <- max(table(forest$observeddata[,Y][1]))/sum(table(forest$observeddata[,Y][1]))
   
+  correct[[toclus + 1]] <- mean(forest$observeddata[,Y][1]== forestpred)
+  correct[[toclus + 2]] <- max(table(forest$observeddata[,Y][1]))/sum(table(forest$observeddata[,Y][1]))
+
   value <- list(medoids = medoids, medoidtrees = mds, clusters=clusters, avgsilwidth=sil, agreement=agreement, accuracy=correct, withinsim=sums, treesimilarities=treesimilarities, treecov=treecov, seed=seed)
   attr(value, "class") <- "clusterforest"
   return(value)
@@ -441,7 +469,7 @@ M3<- function (observeddata, treedata, Y, trees, sameobs){
 }
 
 ###### Measure 4: Partition metric ################
-M4<- function (observeddata, treedata, Y, trees, sameobs){
+M4 <- function (observeddata, treedata, Y, trees, sameobs){
   pamtrees <- lapply(1:length(trees), function (i) pamtree(observeddata,treedata[[i]], Y[i],trees[[i]]))
 
   n <- length(trees)
@@ -452,11 +480,8 @@ M4<- function (observeddata, treedata, Y, trees, sameobs){
   }
 
   par<- lapply(1:n, function (s) part(treedata[[s]], pamtrees[[s]]$prednode))
-  no_cores <- detectCores() - 1
-  cl <- makeCluster(no_cores)
-  clusterExport(cl, c("metr", "par", "treedata", "n"), envir=environment())
-  si <- parSapply(cl, 1:n, function (s) sapply (s:n, function (j) metr(par[[s]], par[[j]], treedata[[s]])))
-  stopCluster(cl)
+
+  si <- sapply(1:n, function (s) sapply (s:n, function (j) metr(par[[s]], par[[j]], treedata[[s]])))
 
   sim <- matrix(0, length(trees), length(trees))
   for (i in 1:n){
@@ -475,7 +500,6 @@ M6 <- function (observeddata, treedata, Y, X, trees, tol){
   pamtrees <- lapply(1:length(trees), function (i) pamtree(observeddata,treedata[[i]], Y[i],trees[[i]]))
 
   n <- length(pamtrees)
-  simmatrix <- matrix(c(0), n, n)
 
   s <- lapply(1:n, function (k) splitvsets(pamtrees[[k]]$path1))
   s0<- lapply(1:n, function (k) splitvsets(pamtrees[[k]]$path0))
@@ -490,16 +514,21 @@ M6 <- function (observeddata, treedata, Y, X, trees, tol){
 
   best<- lapply(1:n, function(k) (sum(treedata[[k]][,Y[k]] == levels(treedata[[k]][,Y[k]])[2])/nrow(treedata[[k]] ))>0.5)
 
-  sim <- sapply(1:n, function (k) sapply(k:n, function (l) simsets(paths1=s[[k]], paths2=s[[l]], paths01=s0[[k]], paths02=s0[[l]], X=X, tol=tol, best1=best[[k]],best2=best[[l]])))
-
-  for (i in 1:n){
-    si<- sim[[i]]
-    si[is.nan(si)] <- 1
-    simmatrix[i, c(i:n)] <- si
+  simmatrix <- foreach (k = 1:n, .combine=cbind) %dopar% {
+    sim <- numeric(n)
+    for (l in k:n) {
+      sim[l] <- simsets(paths1=s[[k]], paths2=s[[l]], paths01=s0[[k]], paths02=s0[[l]], X=X, tol=tol, best1=best[[k]], best2=best[[l]])
+    }
+    sim
   }
 
-  ind <- lower.tri(simmatrix)   #make matrix symmetric
-  simmatrix[ind] <- t(simmatrix)[ind]
+  for (k in 1:n) {
+    for (l in k:n) {
+      simmatrix[k, l] = simmatrix[l, k]
+    }
+  }
+
+  simmatrix = matrix(simmatrix, ncol=length(pamtrees), nrow=length(pamtrees), dimnames = NULL)
   return(simmatrix)
 }
 
@@ -510,19 +539,15 @@ M5 <- function(observeddata, treedata, Y, X, trees, sameobs){
   s <- matrix(0, length(pamtrees), length(pamtrees))
   di <- lapply(1:length(pamtrees), function (i) disjnorm(pamtrees[[i]], trees[[i]],observeddata, treedata[[i]] ,X, Y[i], sameobs))
 
-  no_cores <- detectCores() - 1
-  cl <- makeCluster(no_cores)
-  clusterExport(cl, c("dis", "di", "n", "pamtrees" ), envir=environment())
 
   if(sameobs==TRUE){
-    si <- parSapply(cl, 1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresptrain, pamtrees[[j]]$predresptrain)))
+    si <- sapply(1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresptrain, pamtrees[[j]]$predresptrain)))
   }
   else{
-    si <- parSapply(cl, 1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresp, pamtrees[[j]]$predresp)))
+    si <- sapply(1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresp, pamtrees[[j]]$predresp)))
   }
 
-  si <- parSapply(cl, 1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresp, pamtrees[[j]]$predresp)))
-  stopCluster(cl)
+  si <- sapply(1:n, function (i) sapply (i:n, function (j) dis(di[[i]], di[[j]], pamtrees[[i]]$predresp, pamtrees[[j]]$predresp)))
 
   for (i in 1:n){
     s[i, c(i:n)] <- si[[i]]
@@ -542,54 +567,34 @@ M5 <- function(observeddata, treedata, Y, X, trees, sameobs){
 ## Function to turn each tree into a pam tree, containing the set of rules, the predictions on the full dataset and the nodes
 ## observeddata is the full dataset, treedata is the dataset for the current tree, Y is a vector with the outcome for each tree and tree are the partytrees
 
-pamtree<- function(observeddata,treedata,Y,tree){
+pamtree <- function(observeddata,treedata,Y,tree){
   if(length(tree) > 2){   ## Check whether there was a split
     paths <- listrulesparty(x=tree)   # lists all the paths from root to leave
     prednode <- predict(tree, newdata = observeddata, type = "node")  #predicts node for every row of full data
-    if(is(treedata[,Y], "factor" )) {treedata[,Y] <- as.factor(treedata[,Y])}  #check whether y is a factor, if not-- make it a factor
-    if(is(observeddata[,Y], "factor") ) {observeddata[,Y] <- as.factor(observeddata[,Y])}  #check whether y is a factor, if not-- make it a factor
+    
+    if(!is(treedata[,Y], "factor" )) {treedata[,Y] <- as.factor(treedata[,Y])}  #check whether y is a factor, if not-- make it a factor
+    if(!is(observeddata[,Y], "factor") ) {observeddata[,Y] <- as.factor(observeddata[,Y])}  #check whether y is a factor, if not-- make it a factor
 
-    #if(class(tree)[1]== "glmtree"){
       predresp <- predict(tree, newdata= observeddata, type="response")  #predicts response for every row of full data
-     # predresp[predresp<0.5] <- levels(observeddata[,Y])[1]
-    #  predresp[predresp!=levels(observeddata[,Y])[1]] <- levels(observeddata[,Y])[2]
       predresp <- factor(predresp, levels=c(levels(observeddata[,Y])[1], levels(observeddata[,Y])[2]))
-    #}else{
-     # predresp <- predict(tree, newdata = observeddata, type="prob")  #predicts response for every row of full data
-      #predresp <- predresp[,1]
-      #predresp[predresp<0.5]<- levels(observeddata[,Y])[1]
-      #predresp[predresp!= levels(observeddata[,Y])[1]]<- levels(observeddata[,Y])[2]
-      #predresp <- factor(predresp, levels=c(levels(observeddata[,Y])[1], levels(observeddata[,Y])[2]))
-    #}
 
-
-    #if(class(tree)[1]== "glmtree"){
       predresptrain <- predict(tree, newdata = treedata, type="response")  #predicts response for every row of full data
-     # predresptrain[predresptrain<0.5] <- levels(observeddata[,Y])[1]
-    #  predresptrain[predresptrain!=levels(observeddata[,Y])[1]] <- levels(observeddata[,Y])[2]
       predresptrain <- factor(predresptrain, levels=c(levels(observeddata[,Y])[1], levels(observeddata[,Y])[2]))
-    #}else{
-    #  predresptrain <- predict(tree, newdata = treedata, type="prob")  #predicts response for every row of full data
-    #  predresptrain <- predresptrain[,1]
-    #  predresptrain[predresptrain<0.5]<- levels(observeddata[,Y])[1]
-    #  predresptrain[predresptrain!= levels(observeddata[,Y])[1]]<- levels(observeddata[,Y])[2]
-    #  predresptrain <- factor(predresptrain, levels=c(levels(observeddata[,Y])[1], levels(observeddata[,Y])[2]))
-    #}
+  
+      prednodetrain <- predict(tree, newdata = treedata, type = "node")  #predicts node for every row of tree data
 
-    prednodetrain <- predict(tree, newdata = treedata, type = "node")  #predicts node for every row of tree data
-
-    frame <- matrix(c(0), length(unique(prednodetrain)), 2) #create matrix with one row for each node value
-    frame[, 1] <- sort(unique(prednodetrain))
-    frame[, 2] <- sapply(sort(unique(prednodetrain)), function(k) levels(predresptrain[prednodetrain == k][1])[predresptrain[prednodetrain == k][1]])  # check the predicted response for every node
-
-
-    #the paths that lead to a response of the second level of y
-    path1 <- paths[frame[, 2] == levels(observeddata[,Y])[2]]
-    path0 <- paths[frame[, 2] == levels(observeddata[,Y])[1]]
-
-    paths <- sapply(1:length(paths), function (k) strsplit(paths[k], " & "))  #split rules with multiple conditions in substrings
-    path1 <- sapply(1:length(path1), function (k) strsplit(path1[k], " & "))
-    path0 <- sapply(1:length(path0), function (k) strsplit(path0[k], " & "))
+      frame <- matrix(c(0), length(unique(prednodetrain)), 2) #create matrix with one row for each node value
+      frame[, 1] <- sort(unique(prednodetrain))
+      frame[, 2] <- sapply(sort(unique(prednodetrain)), function(k) levels(predresptrain[prednodetrain == k][1])[predresptrain[prednodetrain == k][1]])  # check the predicted response for every node
+  
+  
+      #the paths that lead to a response of the second level of y
+      path1 <- paths[frame[, 2] == levels(observeddata[,Y])[2]]
+      path0 <- paths[frame[, 2] == levels(observeddata[,Y])[1]]
+  
+      paths <- sapply(1:length(paths), function (k) strsplit(paths[k], " & "))  #split rules with multiple conditions in substrings
+      path1 <- sapply(1:length(path1), function (k) strsplit(path1[k], " & "))
+      path0 <- sapply(1:length(path0), function (k) strsplit(path0[k], " & "))
 
     #if there was no split
   }else{
@@ -1040,17 +1045,26 @@ splitvsets <- function (paths){
 
 simsets<- function (paths1, paths2, paths01, paths02, X, tol,best1,best2){
 
-
-  if(length(paths1[[1]])&length(paths2[[1]])>0) {
-    J1 <- JaccardPaths(paths1, paths2, tol,X)
-    J0 <- JaccardPaths(paths01, paths02,tol,X)
-    sim<- (J1+J0)/2
-
+  if(length(paths1[[1]]) > 0 & length(paths2[[1]]) > 0) { 
+    if(is.null(tol)){
+      if(identical(paths1$splitvars, paths2$splitvars) & identical(paths01$splitvars, paths02$splitvars)){
+        sim <- 1
+      } else {
+        J1 <- JaccardPaths(paths1, paths2, tol,X)
+        J0 <- JaccardPaths(paths01, paths02,tol,X)
+        sim <- (J1+J0)/2
+      }
+    } else {
+      J1 <- JaccardPaths(paths1, paths2, tol,X)
+      J0 <- JaccardPaths(paths01, paths02,tol,X)
+      sim <- (J1+J0)/2
+    }
+    
   } else{
     if(length(paths1[[1]]) == 0 & length(paths2[[1]]) == 0){
       if(best1==best2){
         sim <- 1
-      } else{sim<-0}
+      } else{sim <- 0}
 
     } else{sim <- 0}
   }
@@ -1251,7 +1265,3 @@ disjnorm <- function (pamtree, tree,observeddata, treedata,X, Y, sameobs){
   }
   return(paths)
 }
-
-
-
-
